/*
 * strings.c
 * (c) Copyright 1992 Kevin L. Moore
 * All Rights Reserved
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define DEF_TOKEN_LEN 6
#define COPYRIGHT "(c) Copyright 1992 Kevin L. Moore"
#define PROGRAM   "strings -- V1.20 -- Release Date Feb 02 1992"

/* function prototypes */
void Usage (void) ;
int IsExt (char *szFileName,char *szExt) ;
int IsLzExe (char *szFileName) ;

int main (int argc,char *argv[])
  {
  FILE *hIn ;
  int ch ;
  int TokenLen ;
  char szWkSt[255] = "" ;
  int dCt ;

  if (argc < 2)
    Usage () ;

  /* force filename to uppercase */
  strupr (argv[1]) ;

  /* if file is an 'exe' and compressed by LZEXE then abort */
  if (IsExt (argv[1],"EXE"))
    if ((dCt = IsLzExe (argv[1])) != 0)
      {
      printf ("%s is compressed by LZEXE version %d, unable to extract strings.\n",argv[1],dCt) ;
      exit (EXIT_FAILURE) ;
      }

  /* check for user supplied tokenlen */
  if ((TokenLen = atoi (argv[2])) == 0)
    TokenLen = DEF_TOKEN_LEN ;

  if ((hIn = fopen (argv[1],"rb")) == NULL)
    {
    fprintf (stderr,"Error accessing %s.\n",argv[1]) ;
    exit (EXIT_FAILURE) ;
    }

  dCt = 0 ;
  while ((ch = fgetc (hIn)) != EOF)
    {
    if (ch >= ' ' && ch <= '~')
      szWkSt[dCt++] = (char) ch ;
    else
      {
      szWkSt[dCt] = '\0' ;
      if (strlen (szWkSt) >= TokenLen)
        printf ("%s\n",szWkSt) ;
      dCt = 0 ;
      }
    }

  if (ferror (hIn))
    {
    fprintf (stderr,"Error reading %s\n",argv[1]) ;
    exit (EXIT_FAILURE) ;
    }

  fclose (hIn) ;

  return (EXIT_SUCCESS) ;
  }

/*************
 * Display copyright and how to use program.
 */
void Usage (void)
  {
  fprintf (stderr,"%s\n",PROGRAM) ;
  fprintf (stderr,"%s\n",COPYRIGHT) ;
  fprintf (stderr,"All Rights Reserved\n") ;
  fprintf (stderr,"usage: strings filename [TokenLen]\n\n") ;
  fprintf (stderr,"TokenLen defaults to %d.  This determines how long a printable token\n",DEF_TOKEN_LEN) ;
  fprintf (stderr,"is to be accepted as a string.\n") ;
  fprintf (stderr,"Strings will detect if the file has been compressed by LZEXE, if so will abort.\n") ;
  exit (EXIT_FAILURE) ;
  }

/*************
 * Determines if an filename has a certain extension.
 * returns 1 if true,0 otherwise.
 */
int IsExt (char *szFileName,char *szExt)
  {
  return ((strcmpi ((szFileName + (strlen (szFileName) - 3)),szExt)) == 0) ;
  }

/*************
 * Determines if a file has been compressed by LZEXE
 * Returns version number if true, 0 otherwise
 */
int IsLzExe (char *szFileName)
  {
  unsigned int wHead[0x10] ;
  FILE *hFile ;
  int nVer = 0 ;

  if ((hFile = fopen (szFileName,"rb")) == NULL)
    {
    fprintf (stderr,"Error opening %s\n",szFileName) ;
    exit (EXIT_FAILURE) ;
    }

  fread (wHead,sizeof (wHead[0]),0x10,hFile) ;

  if (wHead[0x0c] == 0x1c && memcmp (&wHead[0x0e],"LZ09",4) == 0)
    nVer = 90 ;

  if (wHead[0x0c] == 0x1c && memcmp (&wHead[0x0e],"LZ91",4) == 0)
    nVer = 91 ;

  fclose (hFile) ;

  return nVer ;
  }


